import numpy as np
from tqdm import tqdm

import torch


Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor

"""
Convention:
 - z: gaussian noise, size 100
 - g(z): fake image generated by generator using noise z
 - x: real image has the same shape as g(z)
"""

# dg/dz
def get_norms_derivative_g_wrt_z(generator, bound, dist, n_samples=5000, latent_dim=100, out_dim=784) -> np.array:
    """
     - Return a numpy array size n_samples
     - Each element is L2 norm of dg/dz
    """
    def get_jacobian(generator, latent_dim, out_dim):
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        z = z.squeeze()
        z = z.repeat(out_dim, 1)
        z.requires_grad_(True)
        y = generator(z)
        t = y.reshape(out_dim, out_dim)
        t.backward(torch.eye(out_dim).cuda())
        return z.grad
    norms = []
    for _ in tqdm(range(n_samples)):
        jacobian_matrix = get_jacobian(generator, latent_dim, out_dim)
        norm = torch.sqrt(jacobian_matrix.pow(2).sum())
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms


# d(d)/dx
def get_norms_derivative_d_wrt_x(real_imgs_loader, discriminator):
    """
     - real_imgs_loader: batch_size = 1 rather than 64 in training
     - Return a numpy array size n_samples
     - Each element is L2 norm of d(d)/dx
    """
    norms = []
    for r_img, _ in tqdm(real_imgs_loader):
        assert r_img.shape[0] == 1

        img = r_img.clone()
        real_img = img.type(Tensor)
        real_img.requires_grad_(True)
        real_pred = discriminator(real_img)
        real_pred = real_pred.squeeze()
        real_pred.backward()
        
        d_wrt_x = real_img.grad
        norm = torch.sqrt(d_wrt_x.pow(2).sum())
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms


# d(d)/d(g(z))
def get_norms_derivative_d_wrt_gz(discriminator, generator, bound, dist, n_samples=5000, latent_dim=100):
    """
     - Return a numpy array size n_samples
     - Each element is L2 norm of d(d)/d(g(z))
    """
    norms = []
    for _ in tqdm(range(n_samples)):
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        with torch.no_grad():
            img = generator(z)
        img.requires_grad_(True)
        fake_img = img.type(Tensor)
        fake_pred = discriminator(fake_img)
        fake_pred = fake_pred.squeeze()
        fake_pred.backward()
        
        d_wrt_gz = img.grad
        norm = torch.sqrt(d_wrt_gz.pow(2).sum())
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms
