import numpy as np
from tqdm import tqdm

import torch


Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor
adversarial_loss = torch.nn.BCELoss()


"""
Convention:
 - z: gaussian noise, size 100
 - g(z): fake image generated by generator using noise z
 - x: real image has the same shape as g(z)
"""


# d(lossd)/dz
def get_norms_derivative_lossd_wrt_z(generator, discriminator, bound, dist, n_samples=5000, latent_dim=100) -> np.array:
    """
     - Return a numpy array size n_samples
     - Each element is L2 norm of d(lossd)/dz
    """
    norms = []
    for _ in tqdm(range(n_samples)):
        fake = Tensor(1, 1).fill_(0.0)
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        z.requires_grad_(True)
        gen_img = generator(z)
        fake_loss = adversarial_loss(discriminator(gen_img), fake) / 2
        fake_loss.backward()

        lossd_wrt_z = z.grad
        norm = torch.sqrt(lossd_wrt_z.pow(2).sum())
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms

def get_norms_derivative_lossg_wrt_z(generator, discriminator, bound, dist, n_samples=5000, latent_dim=100) -> np.array:
    """
     - Return a numpy array size n_samples
     - Each element is L2 norm of d(lossg)/dz
    """
    norms = []
    for _ in tqdm(range(n_samples)):
        valid = Tensor(1, 1).fill_(1.0)
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        z.requires_grad_(True)
        gen_img = generator(z)
        g_loss = -adversarial_loss(valid - discriminator(gen_img), valid)
        g_loss.backward()

        lossg_wrt_z = z.grad
        norm = torch.sqrt(lossg_wrt_z.pow(2).sum())
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms


def get_avg_lossg_lossd_Edx_Edgz_acc(generator, discriminator, real_imgs_loader, bound, dist, latent_dim=100):
    """
     - Return lossg, lossd and acc of d
    """
    lossg = []
    lossd = []
    preds = []
    labels = []
    dx = []
    dgz = []
    for img, _ in tqdm(real_imgs_loader):
        valid = Tensor(1, 1).fill_(1.0)
        fake = Tensor(1, 1).fill_(0.0)
        real_img = img.type(Tensor)
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        with torch.no_grad():
            gen_img = generator(z)
            g_loss = -adversarial_loss(valid - discriminator(gen_img), valid)
        lossg.append(g_loss.item())

        # Measure discriminator's ability to classify real from generated samples
        with torch.no_grad():
            real_pred = discriminator(real_img)
        dx.append(real_pred.squeeze().item())
        real_loss = adversarial_loss(real_pred, valid)
        with torch.no_grad():
            fake_pred = discriminator(gen_img.detach())
        dgz.append(fake_pred.squeeze().item())
        fake_loss = adversarial_loss(fake_pred, fake)

        d_loss = (real_loss + fake_loss) / 2
        lossd.append(d_loss.item())

        preds.append(real_pred.squeeze().item() > 0.5 + 0)
        preds.append(fake_pred.squeeze().item() > 0.5 + 0)
        labels.append(1)
        labels.append(0)

    lossg = np.array(lossg)
    lossd = np.array(lossd)
    dx = np.array(dx)
    dgz = np.array(dgz)

    preds = np.array(preds)
    labels = np.array(labels)
    acc = sum(preds == labels) / len(preds)

    return lossg.mean(), lossg.std(), lossd.mean(), lossd.std(), dx.mean(), dx.std(), dgz.mean(), dgz.std(), acc


def get_norms_derivative_lossg_wrt_theta_g(generator, discriminator, bound, dist, latent_dim=100, n_samples=5000):
    norms = []
    optimizer_G = torch.optim.Adam(generator.parameters())
    optimizer_D = torch.optim.Adam(discriminator.parameters())
    for _ in tqdm(range(n_samples)):
        optimizer_G.zero_grad()
        optimizer_D.zero_grad()
        valid = Tensor(1, 1).fill_(1.0)
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        gen_img = generator(z)
        g_loss = -adversarial_loss(valid - discriminator(gen_img), valid)
        g_loss.backward()
        s = torch.cat([x.grad.flatten() for x in generator.parameters() if x.grad is not None])
        norm = s.norm()
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms


def get_norms_derivative_lossd_wrt_theta_d(generator, discriminator, real_imgs_loader, bound, dist, latent_dim=100):
    optimizer_G = torch.optim.Adam(generator.parameters())
    optimizer_D = torch.optim.Adam(discriminator.parameters())
    norms = []
    for img, _ in tqdm(real_imgs_loader):
        optimizer_G.zero_grad()
        optimizer_D.zero_grad()
        valid = Tensor(1, 1).fill_(1.0)
        fake = Tensor(1, 1).fill_(0.0)
        real_img = img.type(Tensor)
        if dist=='gauss':
            z = Tensor(np.random.normal(0, bound, (1, latent_dim)))
        elif dist=='uniform':
            z = Tensor(np.random.uniform(-bound, bound, (1, latent_dim)))
        else:
            raise RuntimeError(f'{dist} is invalid!')
        with torch.no_grad():
            gen_img = generator(z)

        # Measure discriminator's ability to classify real from generated samples
        real_pred = discriminator(real_img)
        real_loss = adversarial_loss(real_pred, valid)
 
        fake_pred = discriminator(gen_img.detach())
        fake_loss = adversarial_loss(fake_pred, fake)

        d_loss = (real_loss + fake_loss) / 2
        d_loss.backward()
        s = torch.cat([x.grad.flatten() for x in discriminator.parameters() if x.grad is not None])
        norm = s.norm()
        norms.append(norm.cpu().numpy())
    norms = np.array(norms)
    return norms
