import numpy as np

import torch.nn as nn


from src.sn_layer import SpectralNorm

latent_dim = 100
img_shape = (1, 28, 28)


class Generator(nn.Module):
    def __init__(self, use_spec_norm):
        super(Generator, self).__init__()

        def block(in_feat, out_feat, normalize=True):
            if use_spec_norm:
                layers = [SpectralNorm(nn.Linear(in_feat, out_feat))]
            else:
                layers = [nn.Linear(in_feat, out_feat)]
            if normalize:
                layers.append(nn.BatchNorm1d(out_feat, 0.8))
            layers.append(nn.LeakyReLU(0.2, inplace=True))
            return layers

        if use_spec_norm:
            self.model = nn.Sequential(
                *block(latent_dim, 128, normalize=False),
                *block(128, 256, normalize=False),
                *block(256, 512, normalize=False),
                *block(512, 1024, normalize=False),
                SpectralNorm(nn.Linear(1024, int(np.prod(img_shape)))),
                nn.Tanh()
            )
        else:
            self.model = nn.Sequential(
                *block(latent_dim, 128, normalize=False),
                *block(128, 256),
                *block(256, 512),
                *block(512, 1024),
                nn.Linear(1024, int(np.prod(img_shape))),
                nn.Tanh()
            )

    def forward(self, z):
        img = self.model(z)
        img = img.view(img.size(0), *img_shape)
        return img


class Discriminator(nn.Module):
    def __init__(self, use_sigmoid, use_spec_norm):
        super(Discriminator, self).__init__()

        if use_sigmoid:
            if use_spec_norm:
                self.model = nn.Sequential(
                    SpectralNorm(nn.Linear(int(np.prod(img_shape)), 512)),
                    nn.LeakyReLU(0.2, inplace=True),
                    SpectralNorm(nn.Linear(512, 256)),
                    nn.LeakyReLU(0.2, inplace=True),
                    SpectralNorm(nn.Linear(256, 1)),
                    nn.Sigmoid(),
                )
            else:
                self.model = nn.Sequential(
                    nn.Linear(int(np.prod(img_shape)), 512),
                    nn.LeakyReLU(0.2, inplace=True),
                    nn.Linear(512, 256),
                    nn.LeakyReLU(0.2, inplace=True),
                    nn.Linear(256, 1),
                    nn.Sigmoid(),
                )
        else:
            if use_spec_norm:
                self.model = nn.Sequential(
                    SpectralNorm (nn.Linear(int(np.prod(img_shape)), 512)),
                    nn.LeakyReLU(0.2, inplace=True),
                    SpectralNorm(nn.Linear(512, 256)),
                    nn.LeakyReLU(0.2, inplace=True),
                    SpectralNorm(nn.Linear(256, 1)),
                    # nn.Sigmoid(),
                )
            else:
                self.model = nn.Sequential(
                    nn.Linear(int(np.prod(img_shape)), 512),
                    nn.LeakyReLU(0.2, inplace=True),
                    nn.Linear(512, 256),
                    nn.LeakyReLU(0.2, inplace=True),
                    nn.Linear(256, 1),
                    # nn.Sigmoid(),
                )

    def forward(self, img):
        img_flat = img.view(img.size(0), -1)
        validity = self.model(img_flat)

        return validity
