import argparse
import logging
import math
import numpy as np
import os
import pandas as pd
from tqdm import tqdm

import torchvision.transforms as transforms
from torchvision.utils import save_image
from torch.utils.data import DataLoader
from torchvision import datasets
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.functional as F
import torch

from src.models import Generator, Discriminator
from src.utils import get_real_pil_images, get_fake_pil_images, sample_image
from src.fid_score import cal_fid_score
from src.diff_augment import translation
from src.cal_metric_funcs import (
    get_norms_derivative_g_wrt_z,
    get_norms_derivative_d_wrt_gz,
)
from src.cal_metric_funcs_gan1 import (
    get_norms_derivative_lossg_wrt_z,
    get_avg_lossg_lossd_Edx_Edgz_acc,
    get_norms_derivative_lossg_wrt_theta_g,
    get_norms_derivative_lossd_wrt_theta_d
)

parser = argparse.ArgumentParser()
parser.add_argument('--exp_id',type=str, required=True)
parser.add_argument("--shift", type=int, choices=[2, 4, 8, -1])
parser.add_argument("--aug_times", type=int, choices=[0, 5, 15])
parser.add_argument("--interval", type=int, choices=[1, 5, 10])
parser.add_argument("--use_spec_norm_G", type=int, choices=[0, 1])
parser.add_argument("--use_spec_norm_D", type=int, choices=[0, 1])
parser.add_argument("--dist", type=str, choices=['gauss', 'uniform'])
parser.add_argument("--bound", type=float, choices=[0.01, 1, 100])
parser.add_argument("--bound_eval", type=float, default=0.01)
parser.add_argument("--backup_z", type=str, choices=['no', 'half', 'full'])
parser.add_argument("--augmentation", type=int, choices=[0, 1])
parser.add_argument("--n_epochs", type=int)

# unchanged configs
parser.add_argument('--data_path', type=str, default='data/mnist')
parser.add_argument("--batch_size", type=int, default=64, help="size of the batches")
parser.add_argument("--lr", type=float, default=0.0002, help="adam: learning rate")
parser.add_argument("--b1", type=float, default=0.5, help="adam: decay of first order momentum of gradient")
parser.add_argument("--b2", type=float, default=0.999, help="adam: decay of first order momentum of gradient")
parser.add_argument("--n_cpu", type=int, default=8, help="number of cpu threads to use during batch generation")
parser.add_argument("--latent_dim", type=int, default=100, help="dimensionality of the latent space")
parser.add_argument("--img_size", type=int, default=28, help="size of each image dimension")
parser.add_argument("--channels", type=int, default=1, help="number of image channels")
args = parser.parse_args()
exp_name = f'{args.exp_id}_shift_{args.shift}_aug_times_{args.aug_times}_specnorm_G_{args.use_spec_norm_G}_specnorm_D_{args.use_spec_norm_D}_dist_{args.dist}_bound_{args.bound}_bound_eval_{args.bound_eval}_backup_z_{args.backup_z}_aug_{args.augmentation}_n_epochs_{args.n_epochs}'
args.exp_name = exp_name

# make exp_folder
exp_folder = f'experiments/{args.exp_name}'
os.makedirs(exp_folder, exist_ok=False)

logging.basicConfig(filename=f'{exp_folder}/{args.exp_name}.txt',
                    filemode='w',
                    format='%(asctime)s, %(levelname)s: %(message)s',
                    datefmt='%y-%m-%d %H:%M:%S',
                    level=logging.INFO)
console = logging.StreamHandler()
console.setLevel(logging.INFO)
logging.getLogger().addHandler(console)

# logging general information
logging.info(f'\n***************** {args.exp_name.upper()} **************')
for k, v in args._get_kwargs():
    logging.info(f'{k}: {v}')
logging.info('--------------------------------')

logging.info(f'exp_folder: {exp_folder}')
images_folder = os.path.join(exp_folder, 'images')
models_folder = os.path.join(exp_folder, 'models')
os.makedirs(images_folder, exist_ok=True)
os.makedirs(models_folder, exist_ok=True)
logging.info(f'images_folder: {images_folder}')
logging.info(f'models_folder: {models_folder}')

# Loss function
adversarial_loss = torch.nn.BCELoss()

# Initialize generator and discriminator
generator = Generator(use_spec_norm=args.use_spec_norm_G)
discriminator = Discriminator(use_sigmoid=True, use_spec_norm=args.use_spec_norm_D)

logging.info(generator)
logging.info(discriminator)

if torch.cuda.is_available():
    generator.cuda()
    discriminator.cuda()
    adversarial_loss.cuda()

# Configure data loader
os.makedirs(args.data_path, exist_ok=True)
logging.info(f'data at: {args.data_path}')

dataloader = torch.utils.data.DataLoader(
    datasets.MNIST(
        args.data_path,
        train=True,
        download=True,
        transform=transforms.Compose(
            [transforms.Resize(args.img_size), transforms.ToTensor(), transforms.Normalize([0.5], [0.5])]
        ),
    ),
    batch_size=args.batch_size,
    shuffle=True,
)

# for evaluate
real_imgs_loader = torch.utils.data.DataLoader(
    datasets.MNIST(
        args.data_path,
        train=False,
        download=True,
        transform=transforms.Compose(
            [transforms.Resize(args.img_size), transforms.ToTensor(), transforms.Normalize([0.5], [0.5])]
        ),
    ),
    batch_size=1,
    shuffle=False,
)

# Optimizers
optimizer_G = torch.optim.Adam(generator.parameters(), lr=args.lr, betas=(args.b1, args.b2))
optimizer_D = torch.optim.Adam(discriminator.parameters(), lr=args.lr, betas=(args.b1, args.b2))

Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor


# ----------
#  Training
# ----------
if args.backup_z=='no':
    pass
elif args.backup_z=='half':
    n = len(dataloader.dataset)
    logging.info(f'No.real images: {n}')
    if args.dist=='gauss':
        noises = Tensor(np.random.normal(0, args.bound, (n//2 + 1, args.latent_dim)))
    elif args.dist=='uniform':
        noises = Tensor(np.random.uniform(-args.bound, args.bound, (n//2 + 1, args.latent_dim)))
    else:
        raise RuntimeError(f'{args.dist} is invalid')
    logging.info(f'Generated 100% noises: {noises.shape}')
elif args.backup_z=='full':
    n = len(dataloader.dataset)
    logging.info(f'No.real images: {n}')
    if args.dist=='gauss':
        noises = Tensor(np.random.normal(0, args.bound, (n, args.latent_dim)))
    elif args.dist=='uniform':
        noises = Tensor(np.random.uniform(-args.bound, args.bound, (n, args.latent_dim)))
    else:
        raise RuntimeError(f'{args.dist} is invalid')
    logging.info(f'Generated 100% noises: {noises.shape}')
else:
    raise RuntimeError(f'{args.backup_z} is invalid!')

header = None
results = []
for epoch in range(args.n_epochs):
    logging.info(f'epoch: {epoch}')
    generator.train()
    discriminator.train()
    if args.backup_z != 'no':
        noises = noises[torch.randperm(noises.shape[0])]
        batch_counter = 0
    for imgs, _ in tqdm(dataloader):
        batch_size = imgs.shape[0]
        # Adversarial ground truths
        if args.augmentation:
            valid = Tensor(batch_size * (args.aug_times + 1), 1).fill_(1.0)
            fake = Tensor(batch_size * (args.aug_times + 1), 1).fill_(0.0)
        else:
            valid = Tensor(batch_size, 1).fill_(1.0)
            fake = Tensor(batch_size, 1).fill_(0.0)
        # Configure input
        real_imgs = imgs.type(Tensor)
        if args.augmentation:
            real_imgs  = torch.cat([real_imgs] + [translation(real_imgs, shift=args.shift) for _ in range(args.aug_times)])

        # -----------------
        #  Train Generator
        # -----------------

        optimizer_G.zero_grad()

        if args.backup_z=='no':
            # Sample noise as generator input
            if args.dist=='gauss':
                z = Tensor(np.random.normal(0, args.bound, (imgs.shape[0], args.latent_dim)))
            else:
                z = Tensor(np.random.uniform(-args.bound, args.bound, (imgs.shape[0], args.latent_dim)))
        elif args.backup_z=='half':
            if args.dist=='gauss':
                z1 = Tensor(np.random.normal(0, args.bound, (batch_size//2, args.latent_dim)))
            else:
                z1 = Tensor(np.random.uniform(-args.bound, args.bound, (batch_size//2, args.latent_dim)))
            b = batch_size - batch_size//2
            z2 = noises[batch_counter * b:(batch_counter+1)*b, :]
            z = torch.cat((z1, z2))
            z = z[torch.randperm(z.shape[0])]
        else:
            z = noises[batch_counter * batch_size:(batch_counter+1)*batch_size, :]

        # Generate a batch of images
        gen_imgs = generator(z)
        if args.augmentation:
            gen_imgs  = torch.cat([gen_imgs] + [translation(gen_imgs, shift=args.shift) for _ in range(args.aug_times)])

        # Loss measures generator's ability to fool the discriminator
        g_loss = -adversarial_loss(valid - discriminator(gen_imgs), valid)

        g_loss.backward()
        optimizer_G.step()

        # ---------------------
        #  Train Discriminator
        # ---------------------

        optimizer_D.zero_grad()

        # Measure discriminator's ability to classify real from generated samples
        real_pred = discriminator(real_imgs)
        real_loss = adversarial_loss(real_pred, valid)

        fake_pred = discriminator(gen_imgs.detach())
        fake_loss = adversarial_loss(fake_pred, fake)
        d_loss = (real_loss + fake_loss) / 2

        d_loss.backward()
        optimizer_D.step()
        if args.backup_z != 'no':
            batch_counter += 1

    if epoch%args.interval != 0:
        continue

    generator.eval()
    discriminator.eval()

    # save checkpoint
    g_path = os.path.join(models_folder, f'generator_{epoch}.pth')
    d_path = os.path.join(models_folder, f'discriminator_{epoch}.pth')
    torch.save(generator.state_dict(), g_path)
    torch.save(discriminator.state_dict(), d_path)
    logging.info(f'saved g at: {g_path}')
    logging.info(f'saved d at: {d_path}')
    sample_image(images_folder, args.bound_eval, args.dist, generator, epoch)


    # statistic

    entry_0 = [epoch]
    header_0 = ['epoch']

    logging.info('g_wrt_z')
    g_wrt_z = get_norms_derivative_g_wrt_z(generator, args.bound_eval, args.dist)
    entry_1 = [g_wrt_z.mean(), g_wrt_z.std(), g_wrt_z.min(), g_wrt_z.max()]
    header_1 = ['g_wrt_z_mean', 'g_wrt_z_std', 'g_wrt_z_min', 'g_wrt_z_max']
    logging.info(entry_1)

    logging.info('lossg_wrt_theta_g')
    lossg_wrt_theta_g = get_norms_derivative_lossg_wrt_theta_g(generator, discriminator, args.bound_eval, args.dist)
    entry_2 = [lossg_wrt_theta_g.mean(), lossg_wrt_theta_g.std(), lossg_wrt_theta_g.min(), lossg_wrt_theta_g.max()]
    header_2 = ['lossg_wrt_theta_g_mean', 'lossg_wrt_theta_g_std', 'lossg_wrt_theta_g_min', 'lossg_wrt_theta_g_max']
    logging.info(entry_2)

    logging.info('lossd_wrt_theta_d')
    lossd_wrt_theta_d = get_norms_derivative_lossd_wrt_theta_d(generator, discriminator, real_imgs_loader, args.bound_eval, args.dist)
    entry_3 = [lossd_wrt_theta_d.mean(), lossd_wrt_theta_d.std(), lossd_wrt_theta_d.min(), lossd_wrt_theta_d.max()]
    header_3 = ['lossd_wrt_theta_d_mean', 'lossd_wrt_theta_d_std', 'lossd_wrt_theta_d_min', 'lossd_wrt_theta_d_max']
    logging.info(entry_3)

    logging.info('d_wrt_gz')
    d_wrt_gz = get_norms_derivative_d_wrt_gz(discriminator, generator, args.bound_eval, args.dist)
    entry_5 = [d_wrt_gz.mean(), d_wrt_gz.std(), d_wrt_gz.min(), d_wrt_gz.max()]
    header_5 = ['d_wrt_gz_mean', 'd_wrt_gz_std', 'd_wrt_gz_min', 'd_wrt_gz_max']
    logging.info(entry_5)

    logging.info('lossg_wrt_z')
    lossg_wrt_z = get_norms_derivative_lossg_wrt_z(generator, discriminator, args.bound_eval, args.dist)
    entry_6 = [lossg_wrt_z.mean(), lossg_wrt_z.std(), lossg_wrt_z.min(), lossg_wrt_z.max()]
    header_6 = ['lossg_wrt_z_mean', 'lossg_wrt_z_std', 'lossg_wrt_z_min', 'lossg_wrt_zmax']
    logging.info(entry_6)

    logging.info('lossg_mean, lossg_std, lossd_mean, lossd_std, dx_mean, dx_std, dgz_mean, dgz_std, acc')
    lossg_mean, lossg_std, lossd_mean, lossd_std, dx_mean, dx_std, dgz_mean, dgz_std, acc = get_avg_lossg_lossd_Edx_Edgz_acc(generator, discriminator, real_imgs_loader,args.bound_eval, args.dist)
    entry_7 = [lossg_mean, lossg_std, lossd_mean, lossd_std, dx_mean, dx_std, dgz_mean, dgz_std, acc]
    header_7 = ['lossg_mean', 'lossg_std', 'lossd_mean', 'lossd_std', 'dx_mean', 'dx_std', 'dgz_mean', 'dgz_std', 'acc']
    logging.info(entry_7)

    logging.info(f'fid score')
    real_pil_images = get_real_pil_images()
    fake_pil_images = get_fake_pil_images(generator, args.bound_eval, args.dist)
    fid_score = cal_fid_score(real_pil_images, fake_pil_images)
    entry_8 = [fid_score]
    header_8 = ['fid_score']
    logging.info(entry_8)

    entry = entry_0 + entry_1 + entry_2 + entry_3 + entry_5 + entry_6 + entry_7 + entry_8
    if header is None:
        header = header_0 + header_1 + header_2 + header_3 + header_5 + header_6 + header_7 + header_8

    results.append(entry)
    logging.info('--------------------------------')

csv_path = f'{exp_folder}/{args.exp_name}.csv'
df = pd.DataFrame(results, columns=header)
df.to_csv(csv_path, index=None)
logging.info(f'Completed')
