import logging
import math
import numpy as np
from PIL import Image
import random

import torch
import torchvision.transforms as transforms
import torch.nn.functional as F
from torchvision.utils import save_image

Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor


class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def get_real_pil_images(path='data/mnist/MNIST/processed/test.pt'):
    dataset = torch.load(path)
    dataset = dataset[0]
    pil_images = []
    for i in range(dataset.shape[0]):
        pil_image = Image.fromarray(dataset[i, :, :].numpy())
        pil_images.append(pil_image)
    return pil_images


def get_fake_pil_images(generator, bound, dist, n_samples=5000, latent_dim=100, shape=28):
    def norm_ip(img, min, max):
        img.clamp_(min=min, max=max)
        img.add_(-min).div_(max - min + 1e-5)
    pil_images = []
    Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor
    if dist=='gauss':
        Z = Tensor(np.random.normal(0, bound, (n_samples, latent_dim)))
    elif dist=='uniform':
        Z = Tensor(np.random.uniform(-bound, bound, (n_samples, latent_dim)))
    else:
        raise RuntimeError(f'{dist} is invalid!')

    with torch.no_grad():
        imgs = generator(Z)
    norm_ip(imgs, float(imgs.min()), float(imgs.max()))
    arr = imgs.mul(255).add_(0.5).clamp_(0, 255).to('cpu', torch.uint8).numpy()
    for i in range(arr.shape[0]):
        pil_img = Image.fromarray(arr[i, :, :, :].squeeze(), mode='L')
        if  pil_img.size[0] != shape or pil_img.size[1] != shape:
            pil_img = pil_img.resize((shape, shape))
        pil_images.append(pil_img)
    return pil_images


def sample_image(images_folder, bound, dist, generator, epoch, n_row=10, latent_dim=100):
    """Saves a grid of generated digits ranging from 0 to n_classes"""
    # Sample noise
    if dist=='gauss':
        z = Tensor(np.random.normal(0, bound, (n_row ** 2, latent_dim)))
    elif dist=='uniform':
        z = Tensor(np.random.uniform(-bound, bound, (n_row ** 2, latent_dim)))
    else:
        raise RuntimeError(f'{dist} is invalid!')
    with torch.no_grad():
        gen_imgs = generator(z)
    path = f"{images_folder}/{epoch}.png"
    save_image(gen_imgs, path, nrow=n_row, normalize=True)
    logging.info(f'saved image at: {path}')
